//Program: getWeather3
//Author: Niel Markwick (based on work by Alex Sneeringer)
//Last Modified by: Niel Markwick 
//Last Modification Date: 12PM July 6th, 2004
//
// Description: This application is intended to be used in a weather
// module for the SageTV custom stv files.  Currently the program is
// custom made to read the xml from www.weather.com and return the
// data in to a text file in a format specified by mlbdude.
//
// This version uses the DOM library which appears to be installed with
// the latest j2sdk from java.sun.com.  It is a much cleaner
// implementation of the xml scraper than the previous version and am
// laying it out in case there is a desire to use this instead.
//

// Weather.com registration information:
// email: noweather@dodgeit.com
//        check at http://www.dodgeit.com/run/checkmail?mailbox=noweather
// password: NoWeather
//
// SDK info: http://download.weather.com/web/xml/sdk.zip
//

// ----------------------------------------------------------------------
// Thank you for registering for the Weather XML Data Feed.  Please print 
// or save this e-mail message for future reference.
// Please go to the following URL to access the XML Data Feed Software 
// Development Kit (SDK).

// http://download.weather.com/web/xml/sdk.zip

// You may update your user profile or change your product selections at 
// any time by logging in to the weather.com registration system with the 
// following information:

// Sign-In Page: 
// http://registration.weather.com/registration/xmloap/step1
// E-mail address: nielm2001@yahoo.com
// Partner ID: 1005337166
// License Key: b2e9d553aeaa21d4

// If you have any questions about your services, please contact us 
// online.


// Regards,

// Your Friends at weather.com
// Live By It!

// http://www.weather.com

// P.S. - If you did not register for a product or service on weather.com 
// or if you received this message in error, you can automatically 
// deactivate the account now by visiting: 
// http://registration.weather.com/registration/myprofile/step1
// ----------------------------------------------------------------------


import java.io.*;
import javax.xml.parsers.*;
import org.w3c.dom.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;
import org.xml.sax.*;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.Date;

public class getWeather3
{
    // variables for replacing weatherroom with weather.com if specified
    public static String weatherroom_url = "http://www.weatherroom.com/xml/ext/";
    public static String weathercom_url = "http://xoap.weather.com/weather/local/";
    
    // arguments for get (excluding units and days)
    public static String weathercom_args = "?cc=*&prod=xoap&link=xoap";
    public static String partner_id="1005337166";
    public static String licence_key="b2e9d553aeaa21d4";

    // cache duration: 60 minutes
    public static long cache_duration=60*60*1000; //milliseconds

    public static String version="1.1.2";

    // History:
    // 1.0.0 -- 6 Jul 2004 
    //       -- First version based on Alex0230's weatherroom.com extractor
    //
    // 1.0.1 -- 6 Jul 2004 
    //       -- update to use provider info in fulldetails
    // 
    // 1.1.1 -- 7 Jul 2004 
    //       -- update to use add more forecast info (wind, %preciptiation, %humidity)
    //       -- cache's weather.txt for 60 minutes
    //
    // 1.1.2 -- 8 Jul 2004 
    //       -- make Precipitation and Humidity smalled, but add padding to 
    //       -- force word-wrap.
    //       -- Use non-breaking spaces between Precip/Humid: and value

    public static void print_usage() {
	System.out.println("GetWeather3 -- weather.com extractor -- version "+version);
	System.out.println("Usage:");
	System.out.println("   getWeather3 <addr> [outfile] [units] [days] [fulldetails]");
	System.out.println("where:");
	System.out.println("   addr=http://xoap.weather.com/weather/local/LOCATIONID");
	System.out.println("or addr=LOCATIONID (weather.com location code or US zip)"  );
	System.out.println("or addr=file:test.xml (local xml file for debugging purposes)");
	System.out.println("   outfile=weather.txt");
	System.out.println("   units=m for metric units, =s for imperial");
	System.out.println("   days=number of days to forecast (up to 10)");
	System.out.println("   fulldetails -- output all information -- makes incompatible weather.txt");
    }

    public static void main(String[] args) {
	
	//will act as defaults
	String xml_url=new String();
	String output_filename = "Weather.txt";
	String units="m";
	int numdays=5;
	boolean fulldetails=false;
		

	if (args.length <1 || args.length > 5) 
	{
	    print_usage();
	    return;
	}

	// verify argg1 -- URL
	if (args[0].indexOf("http://") != -1 
	    || args[0].indexOf("www.") != -1
	    || args[0].indexOf("file:") != -1) 
	{
	    xml_url = args[0];
	}
	else if ( args[0].indexOf("/") == -1
		  && args[0].indexOf(":") == -1 )
	{
	    // assume location code
	    xml_url = weathercom_url+args[0];
	}
	else
	{
	    System.out.println("invalid URL passed: " + args[0]);
	    print_usage();
	    output_error(output_filename,"invalid Command line","invalid URL");
	    return;
	}
 

	// verify arg2 -- file
	if ( args.length >= 2 )
	{
	    output_filename = args[1];
	}
	 
	// verify arg3 -- units
	if ( args.length >= 3 )
	{
	    if ( args[2].equals("m") ||  args[2].equals("s") ) 
	    {
		units=args[2];
	    }
	    else
	    {
		System.out.println("invalid units: " + args[2]);
		print_usage();
		output_error(output_filename,"invalid Command line","invalid units");
		return;
	    }
	}
	// verify arg4 -- days
	if ( args.length >= 4 )
	{
	    numdays=-1;
	    try {
		numdays=Integer.parseInt(args[3]);
	    } catch (NumberFormatException e) {
	    }
	    if ( numdays < 1 || numdays > 10 ) {
		System.out.println("invalid number of days: " + args[3]);
		print_usage();
		output_error(output_filename,"invalid Command line","invalid number of days");
		return;
	    }
	}

	// verify arg5 -- fulldetails
	if ( args.length >= 5 )	{
	    if ( args[4].equals("fulldetails") ) {
		fulldetails=true;
	    } else {
		System.out.println("invalid 5th argument " + args[4]);
		print_usage();
		output_error(output_filename,"invalid Command line","invalid fulldetails arg");
		return;
	    }
	}

	// attempt to convert weaherroom URL to weather.com URL
	if ( xml_url.indexOf(weatherroom_url) != -1 ) {
	    String zip=xml_url.substring(weatherroom_url.length());
	    xml_url=( weathercom_url 
		      + zip
		      + weathercom_args
		      + "&dayf=" + Integer.toString(numdays)
		      + "&unit="+units
		     + "&par=" + partner_id
		     + "&key=" + licence_key );
	    
	} else if ( xml_url.indexOf("file:") != -1 ) {
	    System.out.println("Warning: using local XML file");

	} else if ( xml_url.indexOf(weathercom_url) != -1 ) {
	    xml_url=(xml_url
		     + weathercom_args
		     + "&dayf=" + Integer.toString(numdays)
		     + "&unit="+units 
		     + "&par=" + partner_id
		     + "&key=" + licence_key);

	} else {
	    System.out.println("Weather Module parser: invalid URL passed");
	    System.out.println(xml_url);
	    print_usage();
	    output_error(output_filename,"invalid Command line","invalid URL");
	    return;
	}

	// attempt to cache the output filename
	// Do this if file is more than cache_duration milliseconds old
	// and has a length greater than 250 bytes
	File weatherfile = new File ( output_filename );
	if ( weatherfile.exists() 
	     && weatherfile.canRead()
	     && weatherfile.lastModified() <=  System.currentTimeMillis() 
	     && ( weatherfile.lastModified() + cache_duration ) > System.currentTimeMillis() 
	     && weatherfile.length() > 250 )
	{
	    System.out.println("Warning: using cached data");
	    return;
	}


//	System.out.println("getting data from:");
//	System.out.println(xml_url);
	getTheWeather(xml_url, output_filename, fulldetails);
    }

    public static void getTheWeather(String xml_url, 
				     String output_filename,
				     boolean fulldetails) {
	
	Document doc = parseXmlFile(xml_url,false);

	try
	{
	    // attempt to catch major errors

	    if (  doc==null ) {
		    output_error(output_filename,
				"Failed to read from weather source",
				"failed to read xml");
		return;
	    }


	    Element weather=(Element)doc.getElementsByTagName("weather").item(0);

	    // check for error result
	    if ( weather==null ) {
		// probably an error doc
		if ( doc.getElementsByTagName("error").getLength()>=1) {
		    output_error(output_filename,
				"Failed to read from weather source",
				"weather.com returned error status: " + getXML(doc.getDocumentElement(), "err", "error"));
		    System.out.println("Weather.com returned error status " + getXML(doc.getDocumentElement(), "err", "error"));	       
		}
		else
		{
		    output_error(output_filename,
				"Failed to read from weather source",
				"weather.com returned empty doc");
		    System.out.println("Weather.com returned empty doc");
		}	       
		return;	    
	    }
	    // got to here, try to parse
	    FileWriter fw = new FileWriter(output_filename);
	    BufferedWriter bw = new BufferedWriter(fw);
	    PrintWriter out = new PrintWriter(bw);	    


	    
	    // attempt to get location info
	    Element head=(Element)weather.getElementsByTagName("head").item(0);
	    // get units
	    // (degree symbol is unicode  00b0)
	    String unit_temp="\u00b0"+getXML(head, "ut", "head");
	    String unit_dist=getXML(head, "ud", "head");
	    String unit_speed=getXML(head, "us", "head");
	    String unit_press=getXML(head, "up", "head");
	    
	    // get location info
	    Element loc=(Element)weather.getElementsByTagName("loc").item(0);
	    
	    
	    // get current forecast info
	    Element cc=(Element)weather.getElementsByTagName("cc").item(0);
	    
	    out.println("curr_location=" + getXML(loc, "dnam", "loc"));
	    out.println("curr_recorded_at="+ getXML(cc, "obst", "cc"));
	    out.println("curr_updated=" + getXML(cc, "lsup", "cc"));
	    out.println("curr_conditions=" + getXML(cc, "t", "cc"));
	    out.println("curr_visibilit=" + getXML(cc, "vis", "cc"));
	    out.println("curr_temp=" + getXML(cc, "tmp", "cc")+unit_temp);
	    out.println("curr_humidity=" + getXML(cc, "hmid", "cc")+"%");
	    out.println("curr_wind=" + ( getXML(cc, "t", "wind") + " " 
					 + getXML(cc, "s", "wind") 
					 + unit_speed ));
	    out.println("curr_dewpoint=" + getXML(cc, "dewp", "cc"));
	    out.println("curr_heatindex=" + getXML(cc, "flik", "cc")+unit_temp);
	    
	    out.println("curr_windchill=" + getXML(cc, "flik", "cc")+unit_temp);
	    out.println("curr_sunrise=" + getXML(loc, "sunr", "loc"));
	    out.println("curr_sunset=" + getXML(loc, "suns", "loc"));
	    out.println("curr_pressure=" + getXML(cc, "r", "bar") +unit_press+ " " + getXML(cc, "d", "bar"));
	    
	    
//          moonphase appears to be unknown	    
//	    out.println("curr_moonphase=");

	    if ( fulldetails ) {
		// print icon, and links 
		out.println("curr_icon=" + getXML(cc, "icon", "cc"));

		out.println("weather_provider=weather.com");
		out.println("weather_provider_url=http://www.weather.com/?prod=xoap&par="+partner_id);
		out.println("weather_provider_img=TWClogo_64px.png");
		


		Element lnks=(Element)weather.getElementsByTagName("lnks").item(0);
		if ( lnks !=null ) {
		    NodeList links=lnks.getElementsByTagName("link");
		    out.println("num_links=" + links.getLength());
		    out.println("links_descr=Featured on weather.com");
		    
		    for ( int linknum=0; linknum<links.getLength(); linknum++) 
		    {
			Element link=(Element)links.item(linknum);
			out.println("link~"+Integer.toString(linknum+1)+"~url=" + getXML(link, "l", "link")+"&par="+partner_id);
			out.println("link~"+Integer.toString(linknum+1)+"~title=" + getXML(link, "t", "link"));
		    }
		}
	    }
    

	    // get future forecasts
	    Element  dayf=(Element)weather.getElementsByTagName("dayf").item(0);
	    NodeList days=dayf.getElementsByTagName("day");
		
	    // do today first

	    
	    Element day=(Element)days.item(0);

	    // Attempt to be clever -- do daytime forecast if time < 17:00, otherwise do nightime
	    String required_part;
	    String required_part_name;
	    GregorianCalendar now = new GregorianCalendar();
	    now.setTime(new Date(System.currentTimeMillis()));
	    if ( now.get(Calendar.HOUR_OF_DAY) < 17 ){
		required_part="d";
		required_part_name="Today";
	    }else{
		required_part="n";
		required_part_name="Tonight";
	    }
	
	    NodeList parts=day.getElementsByTagName("part");
	    for ( int partnum=0; partnum<parts.getLength(); partnum++) 
	    {
		Element part=(Element)parts.item(partnum);
		if ( part.getAttribute("p").equals(required_part) )
		{
		    out.println("~"+required_part_name+"~Conditions=" + getXML(part, "t", "part"));

		    if ( required_part.equals("n") )
		    {
			// only show min temp
			out.println("~"+required_part_name+"~Temp=" + getXML(day, "low", "day")+unit_temp);
		    } else {
			// show min/max on single line
			String max=getXML(day, "hi", "day");
			String min=getXML(day, "low", "day");
			if ( max.equals("N/A") && ! min.equals("N/A") ) {
			    out.println("~"+required_part_name+"~Temp=" + min+unit_temp);

			} else if ( min.equals("N/A")  && ! max.equals("N/A")) {
			    out.println("~"+required_part_name+"~Temp=" + max+unit_temp);
			} else {
			    out.println("~"+required_part_name+"~Temp=" + min+unit_temp+"/"+max+unit_temp);
			}
		    }
		    // Show rest on single line:
		    out.println("~"+required_part_name+"~Wind=" 
				+ getXML(part, "t", "wind") + "\u00A0" + getXML(part, "s", "wind") + unit_speed
				+ "        Precip:\u00A0" + getXML(part, "ppcp", "part") + "%"
				+ "        Humid:\u00A0" + getXML(part, "hmid", "part") + "%");
		    
		    if ( fulldetails) {
			out.println("~"+required_part_name+"~Icon=" + getXML(part, "icon", "part"));
		    }
		}
	    }
	    // then do only days for the remaining forecasts


	    for ( int i=1; i < days.getLength(); i++) 
	    {
		day=(Element)days.item(i);
		String dayname=(day.getAttribute("t"));
		
		parts=day.getElementsByTagName("part");
		for (int  partnum=0; partnum<parts.getLength(); partnum++) 
		{
		    Element part=(Element)parts.item(partnum);
		    if ( part.getAttribute("p").equals("d") )
		    {
			out.println("~"+dayname+"~Conditions=" + getXML(part, "t", "part"));

			// show min/max on single line
			String max=getXML(day, "hi", "day");
			String min=getXML(day, "low", "day");
			if ( max.equals("N/A") && ! min.equals("N/A") ) {
			    out.println("~"+dayname+"~Temp=" + min+unit_temp);
			    
			} else if ( min.equals("N/A")  && ! max.equals("N/A")) {
			    out.println("~"+dayname+"~Temp=" + max+unit_temp);
			} else {
			    out.println("~"+dayname+"~Temp=" + min+unit_temp+"/"+max+unit_temp);
			}
		    
			// Show rest on single line with non-breaking spaces
			// (unicode 00A0)
			out.println("~"+dayname+"~Wind=" 
				    + getXML(part, "t", "wind") + "\u00A0" + getXML(part, "s", "wind") + unit_speed
				    + "        Precip:\u00A0" + getXML(part, "ppcp", "part") + "%"
				    + "        Humid:\u00A0" + getXML(part, "hmid", "part") + "%");
			
			if ( fulldetails) {
			    out.println("~"+dayname+"~Icon=" + getXML(part, "icon", "part"));
			}
		    }
		}
	    }

	    out.close();
	    
	    try
	    {
		bw.close();
		fw.close();
	    }
	    catch (Exception writeClose)
	    {
	    }
	}
	catch (Exception e)
	{
	    System.out.println("Problem parsing XML:");
	    e.printStackTrace(System.out);
	    return ;		
	}
    }
    
    //This method gets the xml out of the document based on the tag and the parent
    //Caveat, it goes off the first match -- limit the matches using element selection
    public static String getXML(Element doc, String tag, String parent) {
	String tmpstr = null;
	try {
	    NodeList tmpNodes = doc.getElementsByTagName(tag);
	    
	    //get the first match
	    for (int i = 0; i < tmpNodes.getLength() && tmpstr==null; i++)
	    {
//		System.out.println(i);
		if ((((Element)(tmpNodes.item(i).getParentNode())).getTagName()).equals(parent))
		{
		    // check children for value
//		    System.out.println("got an  xml node  for "+parent+"/"+tag);
//		    System.out.println("Node: "+((Element)tmpNodes.item(i)).getTagName());
//		    System.out.println("Node text: "+tmpNodes.item(i).toString());

		    NodeList children=tmpNodes.item(i).getChildNodes();
		    for ( int child=0; child < children.getLength(); child++) {
			if ( children.item(child).getNodeType()==Node.TEXT_NODE ) {
//			    System.out.println("value" +children.item(child).getNodeValue());
			    tmpstr=children.item(child).getNodeValue();
			    break;
			}
		    }
		}
	    }
	    
	}
	catch (Exception e) {
	    System.out.println("Failed to get xml value for "+parent+"/"+tag);
	    e.printStackTrace();
	}

	if (tmpstr ==null)
	    // found nothing, return blank
	    return "";

	// clean up string
	
	tmpstr.trim();
	
	//gets rid of the weird (A) character
	if (tmpstr.indexOf((char)194) != -1)
	{
	    tmpstr = tmpstr.substring(0, tmpstr.indexOf((char)194)) + tmpstr.substring(tmpstr.indexOf((char)194) + 1);
	}
	
	//trim didn't do its job
	if ((int)tmpstr.charAt(0) == 32)
	    tmpstr = tmpstr.substring(1);
	
	if ((int)tmpstr.charAt(tmpstr.length() -1) == 32)
	    tmpstr = tmpstr.substring(0, tmpstr.length() - 1);
	
	return tmpstr;
    }

    
    // Source for sample: http://javaalmanac.com/egs/javax.xml.transform/WriteDom.html
    // Parses an XML file and returns a DOM document.
    // If validating is true, the contents is validated against the DTD
    // specified in the file.
    public static Document parseXmlFile(String xml_url, boolean validating) {
	try {
	    // Create a builder factory
	    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
	    factory.setValidating(validating);
	    
	    // Create the builder and parse the file
	    Document doc = factory.newDocumentBuilder().parse(xml_url);
	    return doc;
	} 
	catch (SAXException e) 
	{
	    // A parsing error occurred; the xml input is not valid
	    System.out.println("Error parsing the xml, the input is not valid: " + e);
	    e.printStackTrace();
	} 
	catch (ParserConfigurationException e) 
	{
	    System.out.println("An unexpected ParserConfigurationException occured: " + e);
	    e.printStackTrace();
	} 
	catch (IOException e) 
	{
	    System.out.println("An unexpected IO error occured: " + e);
	    e.printStackTrace();
	}
	return null;
    }
    public static void output_error(String output_filename,String err, String details) {
	try { 
	    FileWriter fw = new FileWriter(output_filename);
	    BufferedWriter bw = new BufferedWriter(fw);
	    PrintWriter out = new PrintWriter(bw);
	    
	    out.println("curr_location="+err);
	    out.println("curr_condition="+details);
	    out.close();
	    bw.close();
	    fw.close();
	}
	catch (Exception e)
	{
	    System.out.println("Failed to write errordoc");	    
	    e.printStackTrace(System.out);
	}
	return;
    }
}
